import copy
from pprint import pformat
from threading import Thread


class MirrorStateStore():
    def __init__(self, shelve_store, dynamodb_store) -> None:
        self.shelve_store = shelve_store
        self.dynamodb_store = dynamodb_store

    def is_valid(self, shelve_kv_pairs: dict, dynamo_kv_pairs: dict, diffs: list) -> bool:
        """
        It checks if all keys in DynamoDB are in shelveStateStore and have the same values.
        It is possible that some keys are in shelveStateStore but is not in DynamoDB when the
        migration starts.
        """
        count = 0
        for k, v in shelve_kv_pairs.items():
            if k in dynamo_kv_pairs.keys():
                count += 1
                dynamo_diff = pformat(dynamo_kv_pairs[k])
                shelve_diff = pformat(shelve_kv_pairs[k])
                if dynamo_diff != shelve_diff:
                    diffs.append(dynamo_diff)
                    diffs.append(shelve_diff)
        return count == len(dynamo_kv_pairs) and not diffs

    def build_key(self, type, iden):
        """
        It builds a unique partition key. The key could be objects with __str__ method.
        """
        return self.shelve_store.build_key(type, iden)

    def save(self, key_value_pairs) -> None:
        """
        Remove all previous data with the same partition key, examine the size of state_data,
        and splice it into different parts under 400KB with different sort keys,
        and save them under the same partition key built.
        """
        key_value_pairs = list(key_value_pairs)
        thread1 = Thread(target = self.shelve_store.save, args=(copy.deepcopy(key_value_pairs),))
        thread2 = Thread(target = self.dynamodb_store.save, args=(key_value_pairs,))
        thread1.start()
        thread2.start()
        thread1.join()
        thread2.join()

    def restore(self, keys) -> dict:
        """
        Fetch all under the same parition key(keys).
        ret: <dict of key to states>
        """
        keys = list(keys)

        def shelve_restore(keys, res):
            res.update(self.shelve_store.restore(keys))

        def dynamodb_restore(keys, res):
            res.update(self.dynamodb_store.restore(keys))

        dynamo_kv_pairs, shelve_kv_pairs = {}, {}
        thread1 = Thread(target = dynamodb_restore, args=(copy.deepcopy(keys), dynamo_kv_pairs,))
        thread2 = Thread(target = shelve_restore, args=(keys, shelve_kv_pairs,))
        thread1.start()
        thread2.start()
        thread1.join()
        thread2.join()

        diffs = []
        if not self.is_valid(shelve_kv_pairs, dynamo_kv_pairs, diffs):
            msg = 'Data in dynamoDB is not synced to BerkleyDB. This is not critical' \
                  + 'since only BerkleyDB is used to restore states right now'
            msg += '\n\n'.join(diffs)
            self.dynamodb_store.alert(msg)

        return shelve_kv_pairs

    def cleanup(self) -> None:
        self.shelve_store.cleanup()



import os
import pickle
import tempfile

import boto3
import pytest
from moto import mock_dynamodb2

from testifycompat import assert_equal
from tron.serialize.runstate.dynamodb_state_store import DynamoDBStateStore


filename = os.path.join(tempfile.mkdtemp(), 'state')


@pytest.fixture
def store():
    with mock_dynamodb2():
        dynamodb = boto3.resource('dynamodb', region_name='us-west-2')
        store = DynamoDBStateStore(filename, 'us-west-2')
        store.table = dynamodb.create_table(
            TableName=filename.replace('/', '-'),
            KeySchema=[
                {
                    'AttributeName': 'key',
                    'KeyType': 'HASH'  # Partition key
                },
                {
                    'AttributeName': 'index',
                    'KeyType': 'RANGE'  # Sort key
                }
            ],
            AttributeDefinitions=[
                {
                    'AttributeName': 'key',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'index',
                    'AttributeType': 'N'
                },

            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': 10,
                'WriteCapacityUnits': 10
            }
        )
        store.client = boto3.client('dynamodb', region_name='us-west-2')
        # Has to be yield here for moto to work
        yield store


@pytest.fixture
def small_object():
    yield pickle.dumps({'this': 'data'})


@pytest.fixture
def large_object():
    yield pickle.dumps([i for i in range(100000)])


@pytest.mark.usefixtures("store", "small_object", "large_object")
class TestDynamoDBStateStore:
    def test_save(self, store, small_object, large_object):
        key_value_pairs = [
            (
                store.build_key("DynamoDBTest", "two"),
                small_object,
            ),
            (
                store.build_key("DynamoDBTest2", "four"),
                small_object,
            ),
        ]
        store.save(key_value_pairs)

        keys = [store.build_key("DynamoDBTest", "two"), store.build_key("DynamoDBTest2", "four")]
        vals = store.restore(keys)
        for key, value in key_value_pairs:
            assert_equal(vals[key], value)

    def test_save_more_than_4KB(self, store, small_object, large_object):
        key_value_pairs = [
            (
                store.build_key("DynamoDBTest", "two"),
                large_object
            )
        ]
        store.save(key_value_pairs)

        keys = [store.build_key("DynamoDBTest", "two")]
        vals = store.restore(keys)
        for key, value in key_value_pairs:
            assert_equal(vals[key], value)

    def test_restore_more_than_4KB(self, store, small_object, large_object):
        keys = [store.build_key("thing", i) for i in range(3)]
        value = pickle.loads(large_object)
        pairs = zip(keys, (value for i in range(len(keys))))
        store.save(pairs)

        vals = store.restore(keys)
        for key in keys:
            assert_equal(pickle.dumps(vals[key]), large_object)

    def test_restore(self, store, small_object, large_object):
        keys = [store.build_key("thing", i) for i in range(3)]
        value = pickle.loads(small_object)
        pairs = zip(keys, (value for i in range(len(keys))))
        store.save(pairs)

        vals = store.restore(keys)
        for key in keys:
            assert_equal(pickle.dumps(vals[key]), small_object)

    def test_delete(self, store, small_object, large_object):
        pairs = [(store.build_key("thing", i), large_object) for i in range(3)]

        store.save(pairs)

        for key in pairs:
            store._delete_item(key)

        for key in pairs:
            assert_equal(store._get_num_of_partitions(key), 0)



import os
import pickle
import tempfile
from pprint import pformat

import boto3
import pytest
from moto import mock_dynamodb2

from testifycompat import assert_equal
from tron.serialize.runstate.dynamodb_state_store import DynamoDBStateStore
from tron.serialize.runstate.mirror_state_store import MirrorStateStore
from tron.serialize.runstate.shelvestore import Py2Shelf
from tron.serialize.runstate.shelvestore import ShelveStateStore


filename = os.path.join(tempfile.mkdtemp(), 'state')


@pytest.fixture
def store():
    with mock_dynamodb2():
        dynamodb = boto3.resource('dynamodb', region_name='us-west-2')
        dynamodb_store = DynamoDBStateStore(filename, 'us-west-2')
        dynamodb_store.table = dynamodb.create_table(
            TableName=filename.replace('/', '-'),
            KeySchema=[
                {
                    'AttributeName': 'key',
                    'KeyType': 'HASH'  # Partition key
                },
                {
                    'AttributeName': 'index',
                    'KeyType': 'RANGE'  # Sort key
                }
            ],
            AttributeDefinitions=[
                {
                    'AttributeName': 'key',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'index',
                    'AttributeType': 'N'
                },

            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': 10,
                'WriteCapacityUnits': 10
            }
        )
        dynamodb_store.client = boto3.client('dynamodb', region_name='us-west-2')
        # Has to be yield here for moto to work
        shelve_store = ShelveStateStore(filename)
        yield MirrorStateStore(shelve_store, dynamodb_store)


@pytest.fixture
def small_object():
    yield pickle.dumps({'this': 'data'})


@pytest.fixture
def large_object():
    yield pickle.dumps([i for i in range(100000)])


@pytest.mark.usefixtures("store", "small_object", "large_object")
class TestDynamoDBStateStore:
    def test_saved_to_both(self, store, small_object, large_object):
        key_value_pairs = [
            (
                store.build_key("DynamoDBTest", "two"),
                large_object
            )
        ]
        # Save to both
        store.save(key_value_pairs)
        # Retreive from both and check if equal
        stored_data = Py2Shelf(filename)
        keys = [store.build_key("DynamoDBTest", "two")]
        vals = store.restore(keys)

        for key, value in key_value_pairs:
            assert_equal(pformat(vals[key]), pformat(value))
            assert_equal(stored_data[str(key.key)], value)
        store.cleanup()

    def test_restore_from_shelve_after_dynamodb_dies(self, store, small_object, large_object):
        key_value_pairs = [
            (
                store.build_key("DynamoDBTest", "two"),
                large_object
            )
        ]
        # Save to both
        store.save(key_value_pairs)
        # This only cleans up data in dynamoDB
        keys = [k for k, v in key_value_pairs]
        # Check if retrievd data is valid
        retrieved_data = store.restore(keys)
        for key in keys:
            assert_equal(retrieved_data[key], large_object)
        store.cleanup()
